#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Taurus Widget qwt5 (former `plot`) module
==========================================


This module provides the Qwt5-based widgets previously provided by
`taurus.qt.qtgui.plot`.
Note that `PyQwt5 module <http://pyqwt.sourceforge.net/>`_
only works with Python2 and PyQt4 and is no longer supported,
so taurus is moving to other modules for plotting (pyqtgraph, silx, ...)
"""

from .qwtdialog import TaurusPlotConfigDialog
from .scales import *
from .taurusplot import *
from .taurustrend import *
from .arrayedit import ArrayEditor
from .taurusarrayedit import TaurusArrayEditor
from .curvesAppearanceChooserDlg import CurveAppearanceProperties, CurvesAppearanceChooser
from .curveprops import CurvePropertiesView
from .monitor import TaurusMonitorTiny
from .curveStatsDlg import CurveStatsDialog
